--------------------------------------------------------------------------------
-- Touhin!
--

local Touhin = LibStub("AceAddon-3.0"):NewAddon("Touhin", "AceEvent-3.0")


--------------------------------------------------------------------------------
-- Locals
--

local media = LibStub("LibSharedMedia-3.0", true)
local deformat = LibStub("LibDeformat-3.0").Deformat

local db = nil
local anchor = nil
local pName = UnitName("player")


--------------------------------------------------------------------------------
-- Options
--

if media then
	media:Register("statusbar", "TouhinDark", [[Interface\AddOns\Touhin\TouhinDark]])
end

local defaults = {
	profile = {
		showMoney = true,
		showLoot = true,
		showTotal = true,
		showRollWinner = true,
		showRollChoices = true,
		qualityThreshold = 3,
		qualitySelfThreshold = 0,
		displayLength = 100,
		growDown = true,
		shownLimit = 10,
		fadeDelay = 8,
		fadeTime = .4,
	}
}

local function GetOptions()
	local options = {
		name = "Touhin",
		type = "group",
		--args = {},
	}
	return options
end


--------------------------------------------------------------------------------
-- Anchor
--

local function CreateAnchor()
	local f = CreateFrame("Frame", nil, UIParent)
	f:SetWidth(220)
	f:SetHeight(18)
	f:ClearAllPoints()
	f:SetPoint(db.anchor_point or "CENTER", db.anchor_x or 0, db.anchor_y or 0)
	
	local texture = f:CreateTexture(nil, "BACKGROUND")
	texture:SetAllPoints(f)
	texture:SetBlendMode("BLEND")
	texture:SetTexture(0, 0, 0, 0.3)
	f.background = texture
	
	local label = f:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
	label:SetText("Touhin")
	label:SetJustifyH("CENTER")
	label:SetJustifyV("MIDDLE")
	label:SetAllPoints(f)
	
	f:EnableMouse(true)
	f:SetClampedToScreen(true)
	f:SetScript("OnMouseUp", function(self, button)
		if button == "RightButton" then
			local point, _, _, x, y = self:GetPoint(1)
			db.anchor_point = point
			db.anchor_x = x
			db.anchor_y = y
			self:Hide()
		elseif button == "LeftButton" and IsShiftKeyDown() then
			Touhin:AddLootTest()
		end
	end)
	
	f:SetMovable(true)
	f:RegisterForDrag("LeftButton")
	f:SetScript("OnDragStart", f.StartMoving)
	f:SetScript("OnDragStop", function(self)
		self:StopMovingOrSizing()
		local point, _, _, x, y = self:GetPoint(1)
		db.anchor_point = point
		db.anchor_x = x
		db.anchor_y = y
	end)
	
	f:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT", 0, 4)
		GameTooltip:ClearLines()
		GameTooltip:AddLine("Touhin") --TODO localize
		GameTooltip:AddLine("|cffeda55fDrag|r to position the anchor.", 0.2, 1, 0.2, 1) --TODO localize
		GameTooltip:AddLine("|cffeda55fRight-Click|r to hide the anchor.", 0.2, 1, 0.2, 1) --TODO localize
		GameTooltip:AddLine("|cffeda55fShift-Click|r to spawn a test frame.", 0.2, 1, 0.2, 1) --TODO localize
		GameTooltip:Show()
	end)
	f:SetScript("OnLeave", function(self)
		GameTooltip:Hide()
	end)
	
	f:Hide()
	
	anchor = f
end

local function ShowAnchor()
	if not anchor then CreateAnchor() end
	anchor:Show()
end

local function HideAnchor()
	if not anchor then return end
	anchor:Hide()
end

local function UpdateAnchor()
	if not anchor then return end
	anchor:ClearAllPoints()
	anchor:SetPoint(db.anchor_point or "CENTER", db.anchor_x or 0, db.anchor_y or 0)
end

local function ResetAnchor()
	if anchor then
		anchor:ClearAllPoints()
		anchor:SetPoint("CENTER")
	end
	db.anchor_point = nil
	db.anchor_x = nil
	db.anchor_y = nil
end


--------------------------------------------------------------------------------
-- Frame stacking
--

local UpdateRows, AddRow, RemoveRow
do
	local function sorter(a, b)
		return a.time < b.time and true or false
	end
	
	local rows = {}
	local tmp = {}
	
	-- FIXME the text seems to "bounce" up one slot when arranging :\ anchoring issue? no idea z.z
	-- removal flow: (Alpha_OnFinished) -> RemoveRow -> Hide -> UpdateRows -> ReleaseFrame
	function UpdateRows()
		wipe(tmp)
		for row in next, rows do
			tmp[#tmp + 1] = row
		end
		table.sort(tmp, sorter)
		
		local last = anchor
		for i=1, db.shownLimit do
			if not tmp[i] then break end
			tmp[i]:ClearAllPoints()
			if db.growDown then
				tmp[i]:SetPoint("TOPLEFT", last, "BOTTOMLEFT", 0, -2)
			else
				tmp[i]:SetPoint("BOTTOMLEFT", last, "TOPLEFT", 0, 2)
			end
			last = tmp[i]
		end
	end

	function AddRow(row)
		if not anchor then CreateAnchor() end
		rows[row] = true
		row:Show()
		UpdateRows()
	end

	function RemoveRow(row)
		row:Hide()
		rows[row] = nil
		UpdateRows()
		--[[--doesn't fix the issue, although it is probably more efficient, lol
		for i=1,#tmp do
			if tmp[i] == row then
				tremove(tmp, i)
				if tmp[i] then
					if db.growDown then
						tmp[i]:SetPoint("TOPLEFT", tmp[i-1] or anchor, "BOTTOMLEFT", 0, -2)
					else
						tmp[i]:SetPoint("BOTTOMLEFT", tmp[i-1] or anchor, "TOPLEFT", 0, 2)
					end
				end
				break
			end
		end]]
	end
end


--------------------------------------------------------------------------------
-- Frame creation
--

local GetFrame, ReleaseFrame
do
	--scripts
	local function frame_OnShow(self)
		self.fader:GetParent():Play()
	end
	
	local function frame_OnEnter(self)
		if not self.link then return end
		GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
		GameTooltip:SetHyperlink(self.link)
	end
	
	local function frame_OnLeave(self)
		if not self.link then return end
		GameTooltip:Hide()
	end
	
	local function frame_OnClick(self, button)
		self.fader:GetParent():Stop()
	end
	
	local function fader_OnFinished(self, requested)
		RemoveRow(self.object)
		ReleaseFrame(self.object)
	end
	
	local backdrop = { --wish i could desaturate the bg just a tiny bit ;[ shit's always too bright
		bgFile = "Interface\\AddOns\\Touhin\\TouhinDark.tga",
		tile = true, tileSize = 32,
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,
		insets = {left = 4, right = 4, top = 4, bottom = 4},
	}
	
	--recycling
	local frameHeap = {}
	
	function GetFrame()
		local frame = tremove(frameHeap)
		--print("GetFrame", frame)
		if not frame then
			--shiney new frame
			frame = CreateFrame("Button", nil, UIParent)
			frame:Hide()
			frame:SetHeight(22)
			frame:SetWidth(220)
			frame:SetBackdrop(backdrop)
			frame:SetBackdropColor(1, 1, 1, .7)
			
			--XXX i should probably do this with one frame, backdrop for border and a texture for the item? hmhm
			local icon = CreateFrame("Button", nil, frame)
			icon:SetPoint("LEFT", 5, 0)
			icon:SetWidth(20)
			icon:SetHeight(20)
			frame.icon = icon
			
			local iconBorder = CreateFrame("Frame", nil, icon)
			iconBorder:SetWidth(26)
			iconBorder:SetHeight(26)
			iconBorder:SetPoint("CENTER", icon, "CENTER")
			iconBorder:SetBackdrop(backdrop)
			iconBorder:SetBackdropColor(1, 1, 1, 0)
			frame.iconBorder = iconBorder
			
			local text = frame:CreateFontString(nil, "OVERLAY", "GameFontWhiteSmall")
			text:SetHeight(16)
			text:SetPoint("LEFT", icon, "RIGHT", 5, 0)
			text:SetPoint("RIGHT", frame, -5, 0)
			text:SetJustifyH("LEFT")
			text:SetJustifyV("MIDDLE")
			text:SetShadowOffset(0.8, -0.8)
			text:SetShadowColor(0, 0, 0, 1)
			frame.text = text
			
			--TODO need/greed/de icon with roll results in a tooltip floating to the right of the frame
			
			--a timer and fancy fading in one, pretty neat
			local ag = frame:CreateAnimationGroup()
			ag:SetLooping("NONE")
			
			local fader = ag:CreateAnimation("Alpha")
			fader:SetChange(-1)
			fader:SetStartDelay(db.fadeDelay)
			--fader:SetEndDelay(0.1)
			fader:SetDuration(db.fadeTime)
			fader:SetSmoothing("IN")
			fader:SetScript("OnStop", fader_OnFinished)
			fader:SetScript("OnFinished", fader_OnFinished)
			fader:SetScript("OnUpdate", nil)
			fader.object = frame
			frame.fader = fader
			
			frame:SetScript("OnShow", frame_OnShow)
			frame:SetScript("OnEnter", frame_OnEnter)
			frame:SetScript("OnLeave", frame_OnLeave)
			frame:SetScript("OnClick", frame_OnClick)
			frame:RegisterForClicks("RightButtonUp")
		else
			frame:SetParent(UIParent)
			frame:SetAlpha(1)
			frame.fader:SetStartDelay(db.fadeDelay)
			frame.fader:SetDuration(db.fadeTime)
		end
		
		return frame
	end

	function ReleaseFrame(frame)
		--print("ReleaseFrame", frame.fader:IsStopped() or frame.fader:IsDone())
		frame:Hide()
		frame:SetParent(nil)
		frame:ClearAllPoints()
		tinsert(frameHeap, frame)
	end
end

--------------------------------------------------------------------------------
-- Initialization
--

function Touhin:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("TouhinDB", defaults, true)
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
	db = self.db.profile
end

function Touhin:OnProfileChanged()
	db = self.db.profile
	UpdateAnchor()
end

function Touhin:OnEnable()
	if not db.anchor_point then
		ShowAnchor()
	end
	
	self:RegisterEvent("CHAT_MSG_LOOT")
	self:RegisterEvent("CHAT_MSG_MONEY")
end


--------------------------------------------------------------------------------
-- Handle Loot!
--

function Touhin:AddLootTest()
	if math.random() > .2 then
		local item
		repeat
			item = GetContainerItemLink(math.random(0, 4), math.random(1, 22))
		until item
		self:AddLoot(pName, item, 1)
	else
		self:AddLoot(pName, "coin", math.random(1, 100000))
	end
end

local function GetCoinString(amount)
	local s = GetCoinText(amount, " ")
	s = string.gsub(s, "(%d+) Gold", "|cffffffff%1|r|cffffd700g|r")
	s = string.gsub(s, "(%d+) Silver", "|cffffffff%1|r|cffc7c7cfs|r")
	s = string.gsub(s, "(%d+) Copper", "|cffffffff%1|r|cffeda55fc|r")
	return s
end

function Touhin:AddLoot(unitName, item, itemCount)
	if not UnitExists(unitName) then return end
	
	local _, class = UnitClass(unitName)
	local itemName, itemLink, itemQuality, itemTexture
	if item ~= "coin" then
		itemName, itemLink, itemQuality, _, _, _, _, _, _, itemTexture = GetItemInfo(item)
		if not itemName then return end
		
		local own = UnitIsUnit("player", unitName)
		if (own and itemQuality < db.qualitySelfThreshold) or (not own and itemQuality < db.qualityThreshold) then
			return
		end
		if own and db.showTotal then
			itemCount = itemCount + GetItemCount(itemLink) or 0
		end
	else
		if not db.showMoney then return end
		itemName = "coin"
		itemQuality = 0
		itemTexture = GetCoinIcon(itemCount)
	end
	
	--print("AddLoot", unitName, itemName, itemCount)
	local row = GetFrame()
	row.link = itemLink
	row.unit = unitName
	
	local qualityColor = ITEM_QUALITY_COLORS[itemQuality]
	
	if item == "coin" then
		row.text:SetText(GetCoinString(itemCount))
		row.time = 0 - itemCount --money always on top (giggity)
	else
		if string.len(itemName) > db.displayLength then
			itemName = string.sub(itemName, 1, db.displayLength) .. ".."
		end
		local classColor = RAID_CLASS_COLORS[class] or {r = 0.8, g = 0.8, b = 0.8}
		classColor = string.format("|cff%02x%02x%02x", classColor.r * 255, classColor.g * 255, classColor.b * 255)
		
		row.text:SetFormattedText("%s%s|r %s[%s]|r |cffaaaaaa%s|r", classColor, unitName, qualityColor.hex, itemName, itemCount)
		row.time = GetTime()
	end
	
	row:SetBackdropColor(qualityColor.r, qualityColor.g, qualityColor.b, .7)
	row:SetBackdropBorderColor(qualityColor.r, qualityColor.g, qualityColor.b, 1)
	row.icon:SetNormalTexture(itemTexture)
	row.iconBorder:SetBackdropBorderColor(qualityColor.r, qualityColor.g, qualityColor.b, 1)
	
	local width = row.text:GetStringWidth()
	row:SetWidth(4 + 5 + 20 + 5 + width + 5) -- insets + pad + icon + pad + text + pad = text + 39
	
	AddRow(row)
end


local function ParseLootMessage(msg)
	local player = pName
	local item, quantity = deformat(msg, LOOT_ITEM_SELF_MULTIPLE)
	if quantity then
		return player, item, tonumber(quantity)
	end
	
	item = deformat(msg, LOOT_ITEM_SELF)
	if item then
		return player, item, 1
	end

	player, item, quantity = deformat(msg, LOOT_ITEM_MULTIPLE)
	if quantity then
		return player, item, tonumber(quantity)
	end
	
	player, item = deformat(msg, LOOT_ITEM)
	if item then
		return player, item, 1
	end
end

function Touhin:CHAT_MSG_LOOT(_, msg)
	local player, item, quantity = ParseLootMessage(msg)
	--print("CHAT_MSG_LOOT", player, item, quantity)
	if player and item and quantity then
		self:AddLoot(player, item, quantity)
	end
end


local ParseMoneyMessage
do
	-- Parse coin strings (code from SpecialEventsEmbed-Loot)
	local cg = GOLD_AMOUNT:gsub("%%d", "(%1+)")
	local cs = SILVER_AMOUNT:gsub("%%d", "(%1+)")
	local cc = COPPER_AMOUNT:gsub("%%d", "(%1+)")
	
	local function parseCoinString(str)
		local g = string.match(str, cg) or 0
		local s = string.match(str, cs) or 0
		local c = string.match(str, cc) or 0
		return g * 10000 + s * 100 + c
	end
	
	function ParseMoneyMessage(msg)
		local player = pName
		local money = deformat(msg, LOOT_MONEY_SPLIT)
		if money then
			return player, parseCoinString(money)
		end
		
		money = deformat(msg, YOU_LOOT_MONEY)
		if money then
			return player, parseCoinString(money)
		end

		player, money = deformat(msg, LOOT_MONEY)
		if player and money then
			return player, parseCoinString(money)
		end
	end
end

function Touhin:CHAT_MSG_MONEY(_, msg)
	local player, quantity = ParseMoneyMessage(msg)
	--print("CHAT_MSG_MONEY", player, quantity)
	if player and quantity then
		self:AddLoot(player, "coin", quantity)
	end
end


SLASH_TOUHIN1 = "/touhin"
SlashCmdList["TOUHIN"] = function(input, editbox)
	ShowAnchor()
end
